/*
 * NPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "NPDA.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

NPDA::NPDA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreAlphabet, label::Label initialState, alphabet::Symbol initialSymbol, std::set < label::Label > finalStates ) : std::Components < NPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

NPDA::NPDA(label::Label initialState, alphabet::Symbol initialPushdownSymbol) : NPDA ( std::set < label::Label > { initialState }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < label::Label > { }) {
}

AutomatonBase* NPDA::clone() const {
	return new NPDA(*this);
}

AutomatonBase* NPDA::plunder() && {
	return new NPDA(std::move(*this));
}

bool NPDA::addTransition(label::Label from, std::variant<string::Epsilon < >, alphabet::Symbol> input, std::vector<alphabet::Symbol> pop, label::Label to, std::vector<alphabet::Symbol> push) {
	if (! getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");

	if (input.is<alphabet::Symbol>() && ! getInputAlphabet().count(input.get<alphabet::Symbol>()))
		throw AutomatonException("Input symbol \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\" doesn't exist.");

	if (! getStates().count(to))
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");

	for(const alphabet::Symbol& popSymbol : pop)
		if (! getPushdownStoreAlphabet().count(popSymbol))
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( popSymbol ) + "\" doesn't exist.");

	for(const alphabet::Symbol& pushSymbol : push)
		if (! getPushdownStoreAlphabet().count(pushSymbol))
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");

	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(std::move(from), std::move(input), std::move(pop));
	std::pair<label::Label, std::vector<alphabet::Symbol> > value = std::make_pair(std::move(to), std::move(push));

	return transitions[std::move(key)].insert(std::move(value)).second;
}

bool NPDA::addTransition(label::Label from, alphabet::Symbol input, std::vector<alphabet::Symbol> pop, label::Label to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool NPDA::addTransition(label::Label from, std::vector<alphabet::Symbol> pop, label::Label to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool NPDA::removeTransition(const label::Label& from, const std::variant<string::Epsilon < >, alphabet::Symbol>& input, const std::vector<alphabet::Symbol>& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(from, input, pop);
	std::pair<label::Label, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);

	return transitions[key].erase(value);
}

bool NPDA::removeTransition(const label::Label& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

bool NPDA::removeTransition(const label::Label& from, const std::vector<alphabet::Symbol>& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return removeTransition(from, inputVariant, pop, to, push);
}

const std::map<std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<label::Label, std::vector<alphabet::Symbol> > > >& NPDA::getTransitions() const {
	return transitions;
}

int NPDA::compare(const NPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void NPDA::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

NPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

NPDA NPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, NPDA::getXmlTagName());

	std::set<label::Label> states = AutomatonFromXMLParser::parseStates<label::Label>(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<alphabet::Symbol>(input);
	std::set<alphabet::Symbol> pushdownStoreSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<alphabet::Symbol>(input);
	label::Label initialState = AutomatonFromXMLParser::parseInitialState<label::Label>(input);
	alphabet::Symbol initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<alphabet::Symbol>(input);
	std::set<label::Label> finalStates = AutomatonFromXMLParser::parseFinalStates<label::Label>(input);

	NPDA automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(pushdownStoreSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<NPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, NPDA::getXmlTagName());
	return automaton;
}

void NPDA::parseTransition(std::deque<sax::Token>::iterator& input, NPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	label::Label from = AutomatonFromXMLParser::parseTransitionFrom<label::Label>(input);
	std::variant<string::Epsilon < >, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<string::Epsilon < >, alphabet::Symbol>(input);
	std::vector<alphabet::Symbol> pop = AutomatonFromXMLParser::parseTransitionPop<alphabet::Symbol>(input);
	label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);
	std::vector<alphabet::Symbol> push = AutomatonFromXMLParser::parseTransitionPush<alphabet::Symbol>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

void NPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(NPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(NPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void NPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionPop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionPush(out, target.second);

			out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto NPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::NPDA>();
auto NPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::NPDA>();

auto NPDACastBinder = castApi::CastPoolStringBinder<automaton::NPDA>(automaton::NPDA::getXmlTagName());

} /* namespace alib */
