/*
 * NPDTA.h
 *
 *  Created on: 10. 5. 2016
 *      Author: Jakub Doupal
 */

#ifndef NPDTA_H_
#define NPDTA_H_

#include "../AutomatonException.h"
#include <set>
#include <map>
#include <vector>
#include <variant>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../../alphabet/Symbol.h"
#include "../../label/Label.h"
#include "../../string/Epsilon.h"

namespace automaton {

class InputAlphabet;
class OutputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Translation Automaton
 */
class NPDTA : public AutomatonBase, public std::Components < NPDTA, alphabet::Symbol, std::tuple < InputAlphabet, OutputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::set < std::tuple < label::Label, std::vector < alphabet::Symbol >, std::vector < alphabet::Symbol > > > > transitions;

public:
	explicit NPDTA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > outputAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, label::Label initialState, alphabet::Symbol initialPushdownSymbol, std::set < label::Label > finalStates );
	explicit NPDTA ( label::Label initialState, alphabet::Symbol initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;


	const label::Label & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( label::Label state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < label::Label > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( label::Label state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < label::Label > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const label::Label & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < label::Label > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( label::Label state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < label::Label > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const label::Label & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getOutputAlphabet ( ) const {
		return accessComponent < OutputAlphabet > ( ).get ( );
	}

	bool addOutputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < OutputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addOutputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < OutputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setOutputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < OutputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeOutputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < OutputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the NPDTA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the NPDTA (state, input symbol, stack symbol, output symbol) or when transition already exists
	 */
	bool addTransition ( label::Label from, std::variant < string::Epsilon < >, alphabet::Symbol > input, std::vector < alphabet::Symbol > pop, label::Label to, std::vector < alphabet::Symbol > push, std::vector < alphabet::Symbol > output );

	bool addTransition ( label::Label from, alphabet::Symbol input, std::vector < alphabet::Symbol > pop, label::Label to, std::vector < alphabet::Symbol > push, std::vector < alphabet::Symbol > output );

	bool addTransition ( label::Label from, std::vector < alphabet::Symbol > pop, label::Label to, std::vector < alphabet::Symbol > push, std::vector < alphabet::Symbol > output );

	/**
	 * Removes the transition from the NPDTA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the NPDTA
	 */
	bool removeTransition ( const label::Label & from, const std::variant < string::Epsilon < >, alphabet::Symbol > & input, const std::vector < alphabet::Symbol > & pop, const label::Label & to, const std::vector < alphabet::Symbol > & push, const std::vector < alphabet::Symbol > & output );

	bool removeTransition ( const label::Label & from, const alphabet::Symbol & input, const std::vector < alphabet::Symbol > & pop, const label::Label & to, const std::vector < alphabet::Symbol > & push, const std::vector < alphabet::Symbol > & output );

	bool removeTransition ( const label::Label & from, const std::vector < alphabet::Symbol > & pop, const label::Label & to, const std::vector < alphabet::Symbol > & push, const std::vector < alphabet::Symbol > & output );

	/**
	 * @return NPDTA transitions
	 */
	const std::map < std::tuple < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::set < std::tuple < label::Label, std::vector < alphabet::Symbol >, std::vector < alphabet::Symbol > > > > & getTransitions ( ) const;

	std::set < std::pair < std::tuple < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::tuple < label::Label, std::vector < alphabet::Symbol >, std::vector < alphabet::Symbol > > > > getTransitionsFromState ( const label::Label & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NPDTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "NPDTA";

		return xmlTagName;
	}

	static NPDTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NPDTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::NPDTA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::NPDTA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> > > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).is<alphabet::Symbol>() && symbol == std::get<1>(transition.first).get<alphabet::Symbol>())
				return true;

		return false;
	}

	static bool available ( const automaton::NPDTA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::NPDTA, alphabet::Symbol, automaton::OutputAlphabet > {
public:
	static bool used ( const automaton::NPDTA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> > > >& transition : automaton.getTransitions())
			for(const std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> >& target : transition.second)
				if (std::find(std::get<2>(target).begin(), std::get<2>(target).end(), symbol) != std::get<2>(target).end())
					return true;

		return false;
	}

	static bool available ( const automaton::NPDTA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::NPDTA, alphabet::Symbol, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::NPDTA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> > > >& transition : automaton.getTransitions()) {
			for (const alphabet::Symbol& popSymbol : std::get<2>(transition.first))
				if (symbol == popSymbol)
					return true;

			for (const std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> >& target : transition.second)
				if (std::find(std::get<1>(target).begin(), std::get<1>(target).end(), symbol) != std::get<1>(target).end())
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDTA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< automaton::NPDTA, alphabet::Symbol, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::NPDTA & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::NPDTA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::NPDTA, label::Label, automaton::States > {
public:
	static bool used ( const automaton::NPDTA & automaton, const label::Label & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> > > >& transition : automaton.getTransitions()) {
			if (state == std::get<0>(transition.first))
				return true;

			for(const std::tuple<label::Label, std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol> >& target : transition.second)
				if(std::get<0>(target) == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::NPDTA &, const label::Label & ) {
		return true;
	}

	static void valid ( const automaton::NPDTA &, const label::Label & ) {
	}
};

template < >
class ComponentConstraint< automaton::NPDTA, label::Label, automaton::FinalStates > {
public:
	static bool used ( const automaton::NPDTA &, const label::Label & ) {
		return false;
	}

	static bool available ( const automaton::NPDTA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDTA &, const label::Label & ) {
	}
};

template < >
class ElementConstraint< automaton::NPDTA, label::Label, automaton::InitialState > {
public:
	static bool available ( const automaton::NPDTA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NPDTA &, const label::Label & ) {
	}
};

} /* namespace std */

#endif /* NPDTA_H_ */
