/*
 * RealTimeHeightDeterministicNPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "RealTimeHeightDeterministicNPDA.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

RealTimeHeightDeterministicNPDA::RealTimeHeightDeterministicNPDA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreAlphabet, std::set < label::Label > initialStates, alphabet::Symbol bottomOfTheStackSymbol, std::set < label::Label > finalStates ) : std::Components < RealTimeHeightDeterministicNPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, label::Label, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > { } ) {
}

RealTimeHeightDeterministicNPDA::RealTimeHeightDeterministicNPDA(alphabet::Symbol bottomOfTheStackSymbol) : RealTimeHeightDeterministicNPDA ( std::set < label::Label > { }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { bottomOfTheStackSymbol }, std::set < label::Label > {}, bottomOfTheStackSymbol, std::set < label::Label > { }) {
}

AutomatonBase* RealTimeHeightDeterministicNPDA::clone() const {
	return new RealTimeHeightDeterministicNPDA(*this);
}

AutomatonBase* RealTimeHeightDeterministicNPDA::plunder() && {
	return new RealTimeHeightDeterministicNPDA(std::move(*this));
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(label::Label from, std::variant<string::Epsilon < >, alphabet::Symbol> input, label::Label to, alphabet::Symbol push) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && ! getInputAlphabet().count(input.get<alphabet::Symbol>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( push ) + "\" doesn't exist.");
	}

	std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>> key(std::move(from), std::move(input));
	std::pair<label::Label, alphabet::Symbol> value = std::make_pair(std::move(to), std::move(push));

	return callTransitions[std::move(key)].insert(std::move(value)).second;
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(label::Label from, label::Label to, alphabet::Symbol push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

bool RealTimeHeightDeterministicNPDA::addCallTransition(label::Label from, alphabet::Symbol input, label::Label to, alphabet::Symbol push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(label::Label from, std::variant<string::Epsilon < >, alphabet::Symbol> input, alphabet::Symbol pop, label::Label to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && !getInputAlphabet().count(input.get<alphabet::Symbol>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));

	return returnTransitions[std::move(key)].insert(std::move(to)).second;
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(label::Label from, alphabet::Symbol pop, label::Label to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

bool RealTimeHeightDeterministicNPDA::addReturnTransition(label::Label from, alphabet::Symbol input, alphabet::Symbol pop, label::Label to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(std::move(input));
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(label::Label from, std::variant<string::Epsilon < >, alphabet::Symbol> input, label::Label to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && ! getInputAlphabet().count(input.get<alphabet::Symbol>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>> key(std::move(from), std::move(input));

	return localTransitions[std::move(key)].insert(std::move(to)).second;
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(label::Label from, label::Label to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool RealTimeHeightDeterministicNPDA::addLocalTransition(label::Label from, alphabet::Symbol input, label::Label to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(std::move(input));
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const label::Label& from, const std::variant<string::Epsilon < >, alphabet::Symbol>& input, const label::Label& to, const alphabet::Symbol& push) {
	std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>> key(from, input);
	std::pair<label::Label, alphabet::Symbol> value = std::make_pair(to, push);

	return callTransitions[key].erase(value);
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const label::Label& from, const label::Label& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::removeCallTransition(const label::Label& from, const alphabet::Symbol& input, const label::Label& to, const alphabet::Symbol& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return removeCallTransition(from, inputVariant, to, push);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const label::Label& from, const std::variant<string::Epsilon < >, alphabet::Symbol>& input, const alphabet::Symbol& pop, const label::Label& to) {
	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const label::Label& from, const alphabet::Symbol& pop, const label::Label& to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::removeReturnTransition(const label::Label& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const label::Label& to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return removeReturnTransition(from, inputVariant, pop, to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const label::Label& from, const std::variant<string::Epsilon < >, alphabet::Symbol>& input, const label::Label& to) {
	std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>> key(from, input);

	return localTransitions[key].erase(to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const label::Label& from, const label::Label& to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return removeLocalTransition(from, inputVariant, to);
}

bool RealTimeHeightDeterministicNPDA::removeLocalTransition(const label::Label& from, const alphabet::Symbol& input, const label::Label& to) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return removeLocalTransition(from, inputVariant, to);
}

const std::map<std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>>, std::set<std::pair<label::Label, alphabet::Symbol> > >& RealTimeHeightDeterministicNPDA::getCallTransitions() const {
	return callTransitions;
}

const std::map<std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::set<label::Label> >& RealTimeHeightDeterministicNPDA::getReturnTransitions() const {
	return returnTransitions;
}

const std::map<std::pair<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>>, std::set<label::Label> >& RealTimeHeightDeterministicNPDA::getLocalTransitions() const {
	return localTransitions;
}

int RealTimeHeightDeterministicNPDA::compare(const RealTimeHeightDeterministicNPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialStates(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialStates(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void RealTimeHeightDeterministicNPDA::operator>>(std::ostream& out) const {
	out << "(RealTimeHeightDeterministicNPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialStates = " << getInitialStates()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

RealTimeHeightDeterministicNPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

RealTimeHeightDeterministicNPDA RealTimeHeightDeterministicNPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());

	std::set<label::Label> states = AutomatonFromXMLParser::parseStates<label::Label>(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<alphabet::Symbol>(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<alphabet::Symbol>(input);
	std::set<label::Label> initialStates = AutomatonFromXMLParser::parseInitialStates<label::Label>(input);
	alphabet::Symbol bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol<alphabet::Symbol>(input);
	std::set<label::Label> finalStates = AutomatonFromXMLParser::parseFinalStates<label::Label>(input);

	RealTimeHeightDeterministicNPDA automaton(std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setInitialStates(std::move(initialStates));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<RealTimeHeightDeterministicNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());
	return automaton;
}

void RealTimeHeightDeterministicNPDA::parseTransition(std::deque<sax::Token>::iterator& input, RealTimeHeightDeterministicNPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		label::Label from = AutomatonFromXMLParser::parseTransitionFrom<label::Label>(input);
		std::variant<string::Epsilon < >, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<string::Epsilon < >, alphabet::Symbol>(input);
		label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);
		alphabet::Symbol push = AutomatonFromXMLParser::parseTransitionSinglePush<alphabet::Symbol>(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		label::Label from = AutomatonFromXMLParser::parseTransitionFrom<label::Label>(input);
		std::variant<string::Epsilon < >, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<string::Epsilon < >, alphabet::Symbol>(input);
		alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop<alphabet::Symbol>(input);
		label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		label::Label from = AutomatonFromXMLParser::parseTransitionFrom<label::Label>(input);
		std::variant<string::Epsilon < >, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<string::Epsilon < >, alphabet::Symbol>(input);
		label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

void RealTimeHeightDeterministicNPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialStates(out, this->getInitialStates());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void RealTimeHeightDeterministicNPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionSinglePush(out, target.second);

			out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getReturnTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getLocalTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto realTimeHeightDeterministicNPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::RealTimeHeightDeterministicNPDA>();
auto realTimeHeightDeterministicNPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::RealTimeHeightDeterministicNPDA>();

auto RealTimeHeightDeterministicNPDACastBinder = castApi::CastPoolStringBinder<automaton::RealTimeHeightDeterministicNPDA>(automaton::RealTimeHeightDeterministicNPDA::getXmlTagName());

} /* namespace alib */
