/*
 * REAL_TIME_HEIGHT_DETERMINISTIC_NPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_
#define REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_

#include "../AutomatonException.h"
#include <map>
#include <vector>
#include <variant>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../../alphabet/Symbol.h"
#include "../../label/Label.h"
#include "../../string/Epsilon.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialStates;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class RealTimeHeightDeterministicNPDA : public AutomatonBase, public std::Components < RealTimeHeightDeterministicNPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, label::Label, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol > >, std::set < std::pair < label::Label, alphabet::Symbol > > > callTransitions;
	std::map < std::tuple < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol >, alphabet::Symbol >, std::set < label::Label > > returnTransitions;
	std::map < std::pair < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol > >, std::set < label::Label > > localTransitions;

public:
	explicit RealTimeHeightDeterministicNPDA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, std::set < label::Label > initialStates, alphabet::Symbol bottomOfTheStackSymbol, std::set < label::Label > finalStates );
	explicit RealTimeHeightDeterministicNPDA ( alphabet::Symbol bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < label::Label > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( label::Label state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < label::Label > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const label::Label & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < label::Label > & getInitialStates ( ) const {
		return accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( label::Label state ) {
		return accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < label::Label > states ) {
		accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	void removeInitialState ( const label::Label & state ) {
		accessComponent < InitialStates > ( ).remove ( state );
	}

	const std::set < label::Label > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( label::Label state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < label::Label > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const label::Label & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getBottomOfTheStackSymbol ( ) const {
		return accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( alphabet::Symbol symbol ) {
		return accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( label::Label current, std::variant < string::Epsilon < >, alphabet::Symbol > input, label::Label next, alphabet::Symbol push );
	bool addCallTransition ( label::Label current, label::Label next, alphabet::Symbol push );
	bool addCallTransition ( label::Label current, alphabet::Symbol input, label::Label next, alphabet::Symbol push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( label::Label current, std::variant < string::Epsilon < >, alphabet::Symbol > input, alphabet::Symbol pop, label::Label next );
	bool addReturnTransition ( label::Label current, alphabet::Symbol pop, label::Label next );
	bool addReturnTransition ( label::Label current, alphabet::Symbol input, alphabet::Symbol pop, label::Label next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( label::Label current, std::variant < string::Epsilon < >, alphabet::Symbol > input, label::Label next );
	bool addLocalTransition ( label::Label current, label::Label next );
	bool addLocalTransition ( label::Label current, alphabet::Symbol input, label::Label next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const label::Label & current, const std::variant < string::Epsilon < >, alphabet::Symbol > & input, const label::Label & next, const alphabet::Symbol & push );
	bool removeCallTransition ( const label::Label & current, const label::Label & next, const alphabet::Symbol & push );
	bool removeCallTransition ( const label::Label & current, const alphabet::Symbol & input, const label::Label & next, const alphabet::Symbol & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const label::Label & current, const std::variant < string::Epsilon < >, alphabet::Symbol > & input, const alphabet::Symbol & pop, const label::Label & next );
	bool removeReturnTransition ( const label::Label & current, const alphabet::Symbol & pop, const label::Label & next );
	bool removeReturnTransition ( const label::Label & current, const alphabet::Symbol & input, const alphabet::Symbol & pop, const label::Label & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const label::Label & current, const std::variant < string::Epsilon < >, alphabet::Symbol > & input, const label::Label & next );
	bool removeLocalTransition ( const label::Label & current, const label::Label & next );
	bool removeLocalTransition ( const label::Label & current, const alphabet::Symbol & input, const label::Label & next );

	const std::map < std::pair < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol > >, std::set < std::pair < label::Label, alphabet::Symbol > > > & getCallTransitions ( ) const;

	const std::map < std::tuple < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol >, alphabet::Symbol >, std::set < label::Label > > & getReturnTransitions ( ) const;

	const std::map < std::pair < label::Label, std::variant < string::Epsilon < >, alphabet::Symbol > >, std::set < label::Label > > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RealTimeHeightDeterministicNPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RealTimeHeightDeterministicNPDA";

		return xmlTagName;
	}

	static RealTimeHeightDeterministicNPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, RealTimeHeightDeterministicNPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::pair<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>>, std::set<std::pair<label::Label, alphabet::Symbol> > >& callTransition : automaton.getCallTransitions())
			if (callTransition.first.second.is<alphabet::Symbol>() && symbol == callTransition.first.second.get<alphabet::Symbol>())
				return true;

		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::set<label::Label> >& returnTransition : automaton.getReturnTransitions())
			if (std::get<1>(returnTransition.first).is<alphabet::Symbol>() && symbol == std::get<1>(returnTransition.first).get<alphabet::Symbol>())
				return true;

		for (const std::pair<const std::pair<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>>, std::set<label::Label> >& localTransition : automaton.getLocalTransitions())
			if (localTransition.first.second.is<alphabet::Symbol>() && symbol == localTransition.first.second.get<alphabet::Symbol>())
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA, alphabet::Symbol, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		for (const std::pair<const std::pair<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>>, std::set<std::pair<label::Label, alphabet::Symbol> > >& callTransition : automaton.getCallTransitions())
			for(const std::pair<label::Label, alphabet::Symbol>& to : callTransition.second)
				if (symbol == to.second)
					return true;

		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::set<label::Label> >& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< automaton::RealTimeHeightDeterministicNPDA, alphabet::Symbol, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA, label::Label, automaton::States > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const label::Label & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>>, std::set<std::pair<label::Label, alphabet::Symbol> > >& callTransition : automaton.getCallTransitions()) {
			if (state == callTransition.first.first)
				return true;

			for(const std::pair<label::Label, alphabet::Symbol>& target : callTransition.second)
				if(target.first == state)
					return true;
		}

		for (const std::pair<const std::tuple<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::set<label::Label> >& returnTransition : automaton.getReturnTransitions()) {
			if (state == std::get<0>(returnTransition.first))
				return true;

			for(const label::Label& target : returnTransition.second)
				if(target == state)
					return true;
		}

		for (const std::pair<const std::pair<label::Label, std::variant<::string::Epsilon < >, alphabet::Symbol>>, std::set<label::Label> >& localTransition : automaton.getLocalTransitions()) {
			if (state == localTransition.first.first)
				return true;

			for(const label::Label& target : localTransition.second)
				if(target == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
	}
};

template < >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA, label::Label, automaton::FinalStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
	}
};

template < >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA, label::Label, automaton::InitialStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA &, const label::Label & ) {
	}
};

} /* namespace std */

#endif /* REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_ */
