/*
 * SinglePopDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "SinglePopDPDA.h"
#include <algorithm>
#include <sstream>
#include <ostream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

SinglePopDPDA::SinglePopDPDA ( std::set < label::Label > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreAlphabet, label::Label initialState, alphabet::Symbol initialSymbol, std::set < label::Label > finalStates ) : std::Components < SinglePopDPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

SinglePopDPDA::SinglePopDPDA(label::Label initialState, alphabet::Symbol initialPushdownSymbol) : SinglePopDPDA ( std::set < label::Label > { initialState }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < label::Label > { }) {
}


AutomatonBase* SinglePopDPDA::clone() const {
	return new SinglePopDPDA(*this);
}

AutomatonBase* SinglePopDPDA::plunder() && {
	return new SinglePopDPDA(std::move(*this));
}

bool SinglePopDPDA::addTransition(label::Label from, std::variant<string::Epsilon < >, alphabet::Symbol> input, alphabet::Symbol pop, label::Label to, std::vector<alphabet::Symbol> push) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.is<alphabet::Symbol>() && ! getInputAlphabet().count(input.get<alphabet::Symbol>())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	for(const alphabet::Symbol& pushSymbol : push) {
		if (!getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));
	std::pair<label::Label, std::vector<alphabet::Symbol> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (input.is<alphabet::Symbol>())
			throw AutomatonException("Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).get<alphabet::Symbol>() ) + "\", \"" + std::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
		else
			throw AutomatonException("Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).get<string::Epsilon < >>() ) + "\", \"" + std::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
	}

	if(input.is<string::Epsilon < >>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::pair<label::Label, std::vector<alphabet::Symbol> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && std::get<2>(transition.first) == std::get<2>(key)) return true;
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol> key2(std::get<0>(key), std::variant<string::Epsilon < >, alphabet::Symbol>{ string::Epsilon < >::EPSILON }, std::get<2>(key));
		if(transitions.find(key2) != transitions.end())
			throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

bool SinglePopDPDA::addTransition(label::Label from, alphabet::Symbol input, alphabet::Symbol pop, label::Label to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool SinglePopDPDA::addTransition(label::Label from, alphabet::Symbol pop, label::Label to, std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

bool SinglePopDPDA::removeTransition(const label::Label& from, const std::variant<string::Epsilon < >, alphabet::Symbol>& input, const alphabet::Symbol& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol> key(from, input, pop);
	std::pair<label::Label, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.is<alphabet::Symbol>())
			throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.get<alphabet::Symbol>() ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");
		else
			throw AutomatonException("Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.get<string::Epsilon < >>() ) + "\") -> \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

bool SinglePopDPDA::removeTransition(const label::Label& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

bool SinglePopDPDA::removeTransition(const label::Label& from, const alphabet::Symbol& pop, const label::Label& to, const std::vector<alphabet::Symbol>& push) {
	std::variant<string::Epsilon < >, alphabet::Symbol> inputVariant(string::Epsilon < >::EPSILON);
	return removeTransition(from, inputVariant, pop, to, push);
}

const std::map<std::tuple<label::Label, std::variant<string::Epsilon < >, alphabet::Symbol>, alphabet::Symbol>, std::pair<label::Label, std::vector<alphabet::Symbol> > >& SinglePopDPDA::getTransitions() const {
	return transitions;
}

int SinglePopDPDA::compare(const SinglePopDPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void SinglePopDPDA::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

SinglePopDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

SinglePopDPDA SinglePopDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, SinglePopDPDA::getXmlTagName());

	std::set<label::Label> states = AutomatonFromXMLParser::parseStates<label::Label>(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<alphabet::Symbol>(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<alphabet::Symbol>(input);
	label::Label initialState = AutomatonFromXMLParser::parseInitialState<label::Label>(input);
	alphabet::Symbol initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<alphabet::Symbol>(input);
	std::set<label::Label> finalStates = AutomatonFromXMLParser::parseFinalStates<label::Label>(input);

	SinglePopDPDA automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<SinglePopDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, SinglePopDPDA::getXmlTagName());
	return automaton;
}

void SinglePopDPDA::parseTransition(std::deque<sax::Token>::iterator& input, SinglePopDPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	label::Label from = AutomatonFromXMLParser::parseTransitionFrom<label::Label>(input);
	std::variant<string::Epsilon < >, alphabet::Symbol> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<string::Epsilon < >, alphabet::Symbol>(input);
	alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop<alphabet::Symbol>(input);
	label::Label to = AutomatonFromXMLParser::parseTransitionTo<label::Label>(input);
	std::vector<alphabet::Symbol> push = AutomatonFromXMLParser::parseTransitionPush<alphabet::Symbol>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

void SinglePopDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void SinglePopDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto singlePopDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::SinglePopDPDA>();
auto singlePopDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::SinglePopDPDA>();

} /* namespace alib */
