/*
 * VisiblyPushdownDPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISIBLY_PUSHDOWN_DPDA_H_
#define VISIBLY_PUSHDOWN_DPDA_H_

#include "../AutomatonException.h"
#include <map>
#include <vector>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../../alphabet/Symbol.h"
#include "../../label/Label.h"

namespace automaton {

class CallAlphabet;
class ReturnAlphabet;
class LocalAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class VisiblyPushdownDPDA : public AutomatonBase, public std::Components < VisiblyPushdownDPDA, alphabet::Symbol, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, label::Label, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < label::Label, alphabet::Symbol >, std::pair < label::Label, alphabet::Symbol > > callTransitions;
	std::map < std::tuple < label::Label, alphabet::Symbol, alphabet::Symbol >, label::Label > returnTransitions;
	std::map < std::pair < label::Label, alphabet::Symbol >, label::Label > localTransitions;

public:
	explicit VisiblyPushdownDPDA ( std::set < label::Label > states, std::set < alphabet::Symbol > callAlphabet, std::set < alphabet::Symbol > returnAlphabet, std::set < alphabet::Symbol > localAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, label::Label initialState, alphabet::Symbol bottomOfTheStackSymbol, std::set < label::Label > finalStates );
	explicit VisiblyPushdownDPDA ( label::Label initialSymbol, alphabet::Symbol bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const label::Label & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( label::Label state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < label::Label > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( label::Label state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < label::Label > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	bool removeState ( const label::Label & state ) {
		return accessComponent < States > ( ).remove ( state );
	}

	const std::set < label::Label > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( label::Label state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < label::Label > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	bool removeFinalState ( const label::Label & state ) {
		return accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getBottomOfTheStackSymbol ( ) const {
		return accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( alphabet::Symbol symbol ) {
		return accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getCallInputAlphabet ( ) const {
		return accessComponent < CallAlphabet > ( ).get ( );
	}

	bool addCallInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < CallAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addCallInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < CallAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setCallInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < CallAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeCallInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < CallAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getReturnInputAlphabet ( ) const {
		return accessComponent < ReturnAlphabet > ( ).get ( );
	}

	bool addReturnInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addReturnInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setReturnInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < ReturnAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeReturnInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < ReturnAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getLocalInputAlphabet ( ) const {
		return accessComponent < LocalAlphabet > ( ).get ( );
	}

	bool addLocalInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < LocalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addLocalInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < LocalAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setLocalInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < LocalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeLocalInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < LocalAlphabet > ( ).remove ( symbol );
	}

	bool removeInputSymbol(const alphabet::Symbol& symbol) {
		return removeCallInputSymbol(symbol) || removeReturnInputSymbol(symbol) || removeLocalInputSymbol(symbol);
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( label::Label current, alphabet::Symbol input, label::Label next, alphabet::Symbol push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( label::Label current, alphabet::Symbol input, alphabet::Symbol pop, label::Label next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( label::Label current, alphabet::Symbol input, label::Label next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const label::Label & current, const alphabet::Symbol & input, const label::Label & next, const alphabet::Symbol & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const label::Label & current, const alphabet::Symbol & input, const alphabet::Symbol & pop, const label::Label & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const label::Label & current, const alphabet::Symbol & input, const label::Label & next );

	const std::map < std::pair < label::Label, alphabet::Symbol >, std::pair < label::Label, alphabet::Symbol > > & getCallTransitions ( ) const;

	const std::map < std::tuple < label::Label, alphabet::Symbol, alphabet::Symbol >, label::Label > & getReturnTransitions ( ) const;

	const std::map < std::pair < label::Label, alphabet::Symbol >, label::Label > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const VisiblyPushdownDPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "VisiblyPushdownDPDA";

		return xmlTagName;
	}

	static VisiblyPushdownDPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, VisiblyPushdownDPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, alphabet::Symbol, automaton::CallAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::pair<label::Label, alphabet::Symbol>, std::pair<label::Label, alphabet::Symbol> >& callTransition : automaton.getCallTransitions())
			if (symbol == callTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in local alphabet");
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in return alphabet");
	}
};

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, alphabet::Symbol, automaton::ReturnAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::tuple<label::Label, alphabet::Symbol, alphabet::Symbol>, label::Label>& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<1>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getLocalInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in local alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in call alphabet");
	}
};

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, alphabet::Symbol, automaton::LocalAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::pair<label::Label, alphabet::Symbol>, label::Label>& localTransition : automaton.getLocalTransitions())
			if (symbol == localTransition.first.second)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getReturnInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in return alphabet");
		if(automaton.getCallInputAlphabet().count(symbol))
			throw automaton::AutomatonException("Input symbol " + (std::string) symbol + " already in call alphabet");
	}
};

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, alphabet::Symbol, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::pair<label::Label, alphabet::Symbol>, std::pair<label::Label, alphabet::Symbol> >& callTransition : automaton.getCallTransitions())
			if (symbol == callTransition.second.second)
				return true;

		for (const std::pair<const std::tuple<label::Label, alphabet::Symbol, alphabet::Symbol>, label::Label>& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< automaton::VisiblyPushdownDPDA, alphabet::Symbol, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::VisiblyPushdownDPDA & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, label::Label, automaton::States > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA & automaton, const label::Label & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<label::Label, alphabet::Symbol>, std::pair<label::Label, alphabet::Symbol> >& callTransition : automaton.getCallTransitions())
			if (state == callTransition.first.first || callTransition.second.first == state)
				return true;

		for (const std::pair<const std::tuple<label::Label, alphabet::Symbol, alphabet::Symbol>, label::Label>& returnTransition : automaton.getReturnTransitions())
			if (state == std::get<0>(returnTransition.first) || returnTransition.second == state)
				return true;

		for (const std::pair<const std::pair<label::Label, alphabet::Symbol>, label::Label>& localTransition : automaton.getLocalTransitions())
			if (state == localTransition.first.first || localTransition.second == state)
				return true;

		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA &, const label::Label & ) {
		return true;
	}

	static void valid ( const automaton::VisiblyPushdownDPDA &, const label::Label & ) {
	}
};

template < >
class ComponentConstraint< automaton::VisiblyPushdownDPDA, label::Label, automaton::FinalStates > {
public:
	static bool used ( const automaton::VisiblyPushdownDPDA &, const label::Label & ) {
		return false;
	}

	static bool available ( const automaton::VisiblyPushdownDPDA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA &, const label::Label & ) {
	}
};

template < >
class ElementConstraint< automaton::VisiblyPushdownDPDA, label::Label, automaton::InitialState > {
public:
	static bool available ( const automaton::VisiblyPushdownDPDA & automaton, const label::Label & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::VisiblyPushdownDPDA &, const label::Label & ) {
	}
};

} /* namespace std */

#endif /* VISIBLY_PUSHDOWN_DPDA_H_ */
